package prometheus

import (
	"testing"
	"time"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	prometheusClient "github.com/prometheus/client_golang/prometheus"
)

func TestPrometheus(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "Prometheus")
}

type testMetric struct {
	value int64
}

func (tm testMetric) Value() int64 {
	return tm.value
}

var _ = Describe("Prometheus Utilities", func() {
	Context("Prometheus Reporter", func() {
		reporter := PrometheusReporter{name: "reporter", statsToUpdateOnScrape: make(map[string]StatMetric), gauges: make(map[string]*prometheusClient.GaugeVec), counters: make(map[string]*prometheusClient.CounterVec), histograms: make(map[string]*prometheusClient.HistogramVec), reporter: prometheusClient.DefaultRegisterer}
		It("has a name", func() {
			Expect(reporter.Name()).To(BeEquivalentTo("reporter"))
		})
		It("handles counters correclty", func() {
			reporter.Count("counter", nil, 1)
			Expect(len(reporter.counters)).To(BeEquivalentTo(1))
			reporter.Count("countertwo", nil, 1)
			Expect(len(reporter.counters)).To(BeEquivalentTo(2))
			counter, err := reporter.getCounter("counter", nil)
			Expect(err).To(BeNil())
			Expect(counter).ToNot(BeNil())
			delete(reporter.counters, "counter")
			counter, err = reporter.getCounter("counter", map[string]string{"newLabel": "foobar"})
			Expect(err).ToNot(BeNil())
			Expect(counter).To(BeNil())
			reporter.counters["counter3"] = nil
			counter, err = reporter.getCounter("counter3", map[string]string{"newLabel": "foobar"})
			Expect(err).To(BeNil())
			Expect(counter).ToNot(BeNil())
		})
		It("handles gauge correctly", func() {
			reporter.Gauge("gauge", nil, 1)
			Expect(len(reporter.gauges)).To(BeEquivalentTo(1))
			reporter.Gauge("gaugetwo", nil, 7)
			Expect(len(reporter.gauges)).To(BeEquivalentTo(2))
			gauge, err := reporter.getGauge("gauge", nil)
			Expect(err).To(BeNil())
			Expect(gauge).ToNot(BeNil())
			delete(reporter.gauges, "gauge")
			gauge, err = reporter.getGauge("gauge", map[string]string{"newLabel": "label"})
			Expect(err).ToNot(BeNil())
			Expect(gauge).To(BeNil())
			reporter.gauges["gauge3"] = nil
			gauge, err = reporter.getGauge("gauge3", map[string]string{"newLabel": "label"})
			Expect(err).To(BeNil())
			Expect(gauge).ToNot(BeNil())
		})
		It("handles histograms correctly", func() {
			reporter.Histogram("hist", nil, 1)
			Expect(len(reporter.histograms)).To(BeEquivalentTo(1))
			reporter.Histogram("histtwo", nil, 1)
			Expect(len(reporter.histograms)).To(BeEquivalentTo(2))
			histogram, err := reporter.getHistogram("hist", nil)
			Expect(err).To(BeNil())
			Expect(histogram).ToNot(BeNil())
			delete(reporter.histograms, "hist")
			histogram, err = reporter.getHistogram("hist", map[string]string{"newLabel": "2"})
			Expect(err).ToNot(BeNil())
			Expect(histogram).To(BeNil())
			reporter.histograms["hist3"] = nil
			histogram, err = reporter.getHistogram("hist3", map[string]string{"newLabel": "2"})
			Expect(err).To(BeNil())
			Expect(histogram).ToNot(BeNil())
		})

		It("tracks method time", func() {
			reporter.TrackMethodTime("method", time.UnixMilli(0), nil)
			tracker, err := reporter.getHistogram("method_duration_micros", nil)
			Expect(err).To(BeNil())
			Expect(tracker).ToNot(BeNil())
		})

		It("Can register scrape value metrics", func() {
			testStat := testMetric{1337}
			err := reporter.RegisterMetric("test", testStat, nil)
			Expect(err).To(BeNil())
			err = reporter.RegisterMetric("test", testStat, nil)
			Expect(err).ToNot(BeNil())
		})

		It("Has a working  factory method", func() {
			Expect(NewReporter()).ToNot(BeNil())
		})
	})
})
