package prometheus

import (
	"fmt"
	"net/http"
	"os"
	"strings"
	"time"

	"hello-api/internal/config"
	"hello-api/internal/logger"

	prometheusClient "github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	"go.uber.org/zap"
	"golang.org/x/exp/maps"
)

const (
	Name                       = "PROMETHEUS"
	prefix                     = "ems_device_hello_"
	errorFindMetricInMap       = "error finding metric in map"
	errorGetMetricFromRegistry = "can't register metric"
)

type PrometheusReporter struct {
	reporter              prometheusClient.Registerer
	counters              map[string]*prometheusClient.CounterVec
	gauges                map[string]*prometheusClient.GaugeVec
	histograms            map[string]*prometheusClient.HistogramVec
	statsToUpdateOnScrape map[string]StatMetric
	name                  string
}

func NewReporter() *PrometheusReporter {
	reporter := &PrometheusReporter{
		reporter:              setupRegistry(),
		name:                  Name,
		gauges:                make(map[string]*prometheusClient.GaugeVec),
		histograms:            make(map[string]*prometheusClient.HistogramVec),
		counters:              make(map[string]*prometheusClient.CounterVec),
		statsToUpdateOnScrape: make(map[string]StatMetric),
	}

	return reporter
}

func setupRegistry() prometheusClient.Registerer {
	registry := prometheusClient.WrapRegistererWithPrefix(prefix, prometheusClient.DefaultRegisterer)
	registry = prometheusClient.WrapRegistererWith(map[string]string{
		"env": config.GetInstance().BaseDimensions["env"],
		"source": func() string {
			hostname, err := os.Hostname()
			if err != nil {
				hostname = "unknown"
			}

			return hostname
		}(),
	}, registry)

	prometheusClient.DefaultRegisterer = registry
	return registry
}

type StatMetric interface {
	Value() int64
}

func (r *PrometheusReporter) StartServer() {
	cfg := config.GetInstance()

	logger.BaseLogger().Info(fmt.Sprintf("Prometheus scrape endpoint on port %d", cfg.PrometheusPort))

	server := &http.Server{
		Addr:              fmt.Sprintf(":%d", cfg.PrometheusPort),
		ReadHeaderTimeout: 3 * time.Second,
	}

	http.Handle("/"+cfg.ServicePrefix+"/prometheus", r.statMetricUpdateMiddleware(promhttp.Handler()))

	go func() {
		err := server.ListenAndServe()
		if err != nil {
			logger.BaseLogger().Error(fmt.Sprintf("Unable to start prometheus scrape endpoint %e", err))
		}
	}()
}

func (r *PrometheusReporter) statMetricUpdateMiddleware(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, request *http.Request) {
		for name, m := range r.statsToUpdateOnScrape {
			r.Gauge(name, nil, m.Value())
		}

		next.ServeHTTP(w, request)
	})
}

func (r *PrometheusReporter) RegisterMetric(name string, metric interface{}, tags map[string]string) error {
	prometheusCompliantName := strings.ReplaceAll(name, ".", "_")

	if _, ok := r.statsToUpdateOnScrape[prometheusCompliantName]; ok {
		logger.BaseLogger().Warn(fmt.Sprintf("Dropping registration of metric %s as it has already been registered", prometheusCompliantName))
		return fmt.Errorf("metric already registered")
	}

	// This is currently used by db_stat_metric, redis_stat_metric, and go_stat_metric which all use Gauges
	// We'll have to account for other metricTypes if we introduce a stat metric that uses another type
	m := prometheusClient.NewGaugeVec(prometheusClient.GaugeOpts{
		Name: prometheusCompliantName,
	}, nil)
	r.statsToUpdateOnScrape[prometheusCompliantName] = metric.(StatMetric)
	r.gauges[prometheusCompliantName] = m
	return r.reporter.Register(m)
}

func (r *PrometheusReporter) TrackMethodTime(name string, start time.Time, dims map[string]string) {
	r.Histogram(
		name+"_duration_micros",
		dims,
		time.Since(start).Nanoseconds()/1e3,
	)
}

func (r *PrometheusReporter) Count(name string, dimensions map[string]string, value int64) {
	metric, err := r.getCounter(name, dimensions)
	if err != nil {
		logger.BaseLogger().Error("get counter failed", zap.String("metric", name), zap.Error(err))
		return
	}

	counterVec, err := metric.GetMetricWith(dimensions)
	if err != nil {
		logger.BaseLogger().Error("unable to get counter with dimensions", zap.String("name", name), zap.Error(err))
		return
	}
	if value >= 0 {
		counterVec.Add(float64(value))
	}
}

func (r *PrometheusReporter) Gauge(name string, dimensions map[string]string, value int64) {
	metric, err := r.getGauge(name, dimensions)
	if err != nil {
		logger.BaseLogger().Error("get gauge failed", zap.String("metric", name), zap.Error(err))
		return
	}

	gaugeVec, err := metric.GetMetricWith(dimensions)
	if err != nil {
		logger.BaseLogger().Error("unable to get gauge with dimensions", zap.String("name", name), zap.Error(err))
		return
	}
	gaugeVec.Set(float64(value))
}

func (r *PrometheusReporter) Histogram(name string, dimensions map[string]string, value int64) {
	metric, err := r.getHistogram(name, dimensions)
	if err != nil {
		logger.BaseLogger().Error("get histogram failed", zap.String("metric", name), zap.Error(err))
		return
	}

	histogramVec, err := metric.GetMetricWith(dimensions)
	if err != nil {
		logger.BaseLogger().Error("unable to get histogram with dimensions", zap.String("name", name), zap.Error(err))
		return
	}
	histogramVec.Observe(float64(value))
}

func (r *PrometheusReporter) Name() string {
	return r.name
}

func (r *PrometheusReporter) getGauge(name string, dimensions map[string]string) (*prometheusClient.GaugeVec, error) {
	prometheusCompliantName := strings.ReplaceAll(name, ".", "_")
	if metric, ok := r.gauges[prometheusCompliantName]; ok && metric != nil {
		return metric, nil
	}

	gauge := prometheusClient.NewGaugeVec(prometheusClient.GaugeOpts{
		Name: prometheusCompliantName,
	},
		maps.Keys(dimensions))

	err := r.reporter.Register(gauge)
	if err != nil {
		return nil, err
	}
	r.gauges[prometheusCompliantName] = gauge
	return gauge, nil
}

func (r *PrometheusReporter) getCounter(name string, dimensions map[string]string) (*prometheusClient.CounterVec, error) {
	prometheusCompliantName := strings.ReplaceAll(name, ".", "_")
	if metric, ok := r.counters[prometheusCompliantName]; ok && metric != nil {
		return metric, nil
	}

	gauge := prometheusClient.NewCounterVec(prometheusClient.CounterOpts{
		Name: prometheusCompliantName,
	},
		maps.Keys(dimensions))

	err := r.reporter.Register(gauge)
	if err != nil {
		return nil, err
	}
	r.counters[prometheusCompliantName] = gauge
	return gauge, nil
}

func (r *PrometheusReporter) getHistogram(name string, dimensions map[string]string) (*prometheusClient.HistogramVec, error) {
	prometheusCompliantName := strings.ReplaceAll(name, ".", "_")
	if metric, ok := r.histograms[prometheusCompliantName]; ok && metric != nil {
		return metric, nil
	}

	histogram := prometheusClient.NewHistogramVec(prometheusClient.HistogramOpts{
		Name: prometheusCompliantName,
	},
		maps.Keys(dimensions))

	err := r.reporter.Register(histogram)
	if err != nil {
		return nil, err
	}
	r.histograms[prometheusCompliantName] = histogram
	return histogram, nil
}
